# Copyright 2015 Greg M. Bernstein
"""
This is an outline of a general POX application that works with ZeroMQ to allow an SDN
application to be split amongst multiple communicating entities.
This program sets up two different ZeroMQ end points. (1) A Request end point to run various methods
in the controller, and (2) a Publisher end point to let client know about application events.
The functionality is similar to that demonstrated with the code OutlineApp.py enhanced with a
messaging interface.

The Request end point runs in a separate thread (see the end of the file for implementation),
while the Publish end point runs in this thread.

This application will only work with Mininet networks that are loop free since it uses flooding.

One needs to set the Python path variables so python/pox can find this module and its dependencies. From
the directory where this file exists I use the following (MS Windows):
set PYTHONPATH=.;..;
python \pox\pox.py py OutlineAppZMQ

To launch mininet using a remote controller:
sudo mn --topo linear,3,2 --controller=remote,ip=192.168.1.134


We can receive status updates outside of POX by running the sub_client.py program and
one can send commands to the running OutlineAppZMQ by running code snippets extracted from
the file send_ex.py.

"""

from pox.core import core
import pox.openflow.libopenflow_01 as of
import pox.lib.packet as pkt
import zmq  # Here we get ZeroMQ
import threading

from collections import deque  # Standard python queue structure

log = core.getLogger()
# Setup the ZeroMQ endpoint URL.
PUB_URL = "tcp://*:5555"  # Used for publishing information
REQ_URL = "tcp://*:5556"  # Used to receive requests


class OutlineAppZMQ(object):
    """
    Waits for OpenFlow switches to connect and then fills their forwarding tables based on
    given network topology.
    """
    def __init__(self):
        """
            Define member variables here and register the handler functions to receive open flow events.
            You do not need to keep these various up and down events.  I do this here so that we can
            look at them via the POX python command line.  I use the standard Python deque data structure
            with a finite length so that I only keep the latest few events of a given type.
        """
        context = zmq.Context.instance()
        self.socket = context.socket(zmq.PUB)  # Set up the ZMQ publisher "socket"
        self.socket.bind(PUB_URL)
        self.switches = {}
        self.down_events = deque(maxlen=10)
        self.packet_in_events = deque(maxlen=10)
        self.port_events = deque(maxlen=10)
        # Let's see what kinds of packets get sent to the controller by default...
        self.packet_in_types = {pkt.ethernet.ARP_TYPE: 0, pkt.ethernet.IP_TYPE: 0, pkt.ethernet.IPV6_TYPE: 0}
        core.openflow.addListeners(self)

    def _handle_ConnectionUp(self, event):
        """ This method gets called when a switch connects to the controller.
        """
        connection = event.connection
        print "Switch {} came up".format(connection.dpid)
        self.socket.send_string("Switch {} came up".format(connection.dpid))  # Publish switch up message
        print "Port information {}".format(connection.ports)
        self.switches[connection.dpid] = connection  # Save switch information

    def _handle_ConnectionDown(self, event):
        """ This is where you would handle a controller switch connectivity down event """
        print "I heard a switch {} connection go down...".format(event.dpid)
        self.socket.send_string("I heard a switch {} connection go down...".format(event.dpid))  # Publish
        self.down_events.appendleft(event)

    def _handle_PacketIn(self, event):
        self.packet_in_events.append(event)
        # Let's keep some counts on the types of packets received
        packet = event.parsed
        ptype = packet.type
        if ptype not in self.packet_in_types:
            self.packet_in_types[ptype] = 0
        self.packet_in_types[ptype] += 1

        # Now let's see if we intercepted a UDP packet
        if ptype != pkt.ethernet.IP_TYPE:
            return
        ip_packet = packet.payload
        tmp = "Intercepted a packet from: {} to {}".format(ip_packet.srcip, ip_packet.dstip)
        print tmp
        self.socket.send_string(tmp)
        if ip_packet.protocol != pkt.ipv4.UDP_PROTOCOL:
            return
        udp_packet = ip_packet.payload
        tmp = "Intercepted UDP packet with source port {} and dest port {}".format(udp_packet.srcport,
                                                                                   udp_packet.dstport)
        print tmp
        self.socket.send_string(tmp)  # Publish UDP packet intercept message.

    def _handle_PortStatus(self, event):
        """Gather up some information about the event to print"""
        switch = event.dpid
        port = event.port
        print "I heard a port status change from switch {} port {}".format(switch, port)
        ofp_phy_port = event.ofp.desc
        name = ofp_phy_port.name
        if  ofp_phy_port.state & 1 == 0:
            state = "Up"
        else:
            state = "Down"
        print "Port {} is {}".format(name, state)
        self.socket.send_string("Port {} is {}".format(name, state))  # Publish port status message
        self.port_events.appendleft(event)

    def make_dumb_hubs(self):
        """ A command to turn wonderful OpenFlow enabled switches into dumb hubs"""
        msg = of.ofp_flow_mod()
        msg.priority = 1
        msg.actions.append(of.ofp_action_output(port=of.ofp_port_rev_map['OFPP_FLOOD']))
        for switch in self.switches.keys():  # Send the flow message to all switches
            connection = self.switches[switch]
            connection.send(msg)
        self.socket.send_string("Somebody made a bunch of dumb hubs!")  # Publish status

    def clear_dumb_hubs(self):
        """ Clears table entries"""
        msg = of.ofp_flow_mod(command=of.ofp_flow_mod_command_rev_map["OFPFC_DELETE"])
        for switch in self.switches.keys():  # Send the flow message to all switches
            connection = self.switches[switch]
            connection.send(msg)
        self.socket.send_string("Somebody cleared the switch entries.")  # Publish status

    def udp_intercept(self, udp_port):
        # Set up a redirect for a particular UDP port
        # Match on Ethertype = 0x0800 IP, UDP is IP protocol number 17
        print "udp_port argument = {}".format(udp_port)
        msg = of.ofp_flow_mod()
        msg.priority = 20
        udp_match = of.ofp_match(dl_type=0x800, nw_proto=17, tp_dst=udp_port)
        msg.match = udp_match
        msg.actions.append(of.ofp_action_output(port=0xfffd))  # This is the controller port
        for switch in self.switches.keys():  # Send the flow message to all switches
            connection = self.switches[switch]
            connection.send(msg)
        self.socket.send_string("Somebody set up a UDP intercept.")  # Publish status


def request_dispatcher():
    """ My request processing routine that receives ZeroMQ request messages and then runs
        an appropriate method via the POX callLater(...) method.  Messages are sent via JSON/Python
        objects and have the form {"type": "XXX", type specific data}  where type can have the string
        values: MakeDumbHubs, ClearDumbHubs, UDPIntercept
    """
    print "Starting request dispatcher..."
    context = zmq.Context.instance()
    socket = context.socket(zmq.REP)
    socket.bind(REQ_URL)

    while True:
        req_obj = socket.recv_json()
        print("Received request in dispatcher")
        if "type" not in req_obj:
            socket.send_json({"type": "Error", "what": "Unknown message, no 'type' field."})
            continue
        app = core.OutlineAppZMQ
        type = req_obj["type"]
        if type == "MakeDumbHubs":
            # Need to communicate with POX's main loop so we use the callLater() from POX core.py
            core.callLater(app.make_dumb_hubs)
            socket.send_json({"type": "processing", "what": req_obj})
        elif type == "ClearDumbHubs":
            core.callLater(app.clear_dumb_hubs)
            socket.send_json({"type": "processing", "what": req_obj})
        elif type == "UDPIntercept":
            port = req_obj["port"] # Gets the UDP port parameter
            core.callLater(app.udp_intercept, port)
            socket.send_json({"type": "processing", "what": req_obj})
        else:
            socket.send_json({"type": "Error", "what": "Unknown message type."})



def launch():
    """
    Starts our OutlineAppZMQ application in POX.
    """
    # Spin up worker thread here for our ZeroMQ reply server.
    thread = threading.Thread(target=request_dispatcher)
    thread.daemon = True
    thread.start()
    core.registerNew(OutlineAppZMQ)
